package com.sun.jndi.dns;

import javax.naming.ConfigurationException;
import javax.naming.directory.Attribute;
import javax.naming.directory.BasicAttribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.InvalidNameException;
import javax.naming.CompositeName;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import javax.naming.directory.ModificationItem;
import javax.naming.directory.DirContext;
import javax.naming.directory.Attributes;
import javax.naming.Context;
import javax.naming.OperationNotSupportedException;
import javax.naming.Binding;
import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.spi.DirectoryManager;
import com.sun.jndi.toolkit.ctx.Continuation;
import javax.naming.Name;
import javax.naming.directory.InvalidAttributeIdentifierException;
import javax.naming.NamingException;
import javax.naming.NameParser;
import java.util.Hashtable;
import com.sun.jndi.toolkit.ctx.ComponentDirContext;

public class DnsContext extends ComponentDirContext
{
    DnsName domain;
    Hashtable<Object, Object> environment;
    private boolean envShared;
    private boolean parentIsDns;
    private String[] servers;
    private com.sun.jndi.dns.Resolver resolver;
    private boolean authoritative;
    private boolean recursion;
    private int timeout;
    private int retries;
    static final NameParser nameParser;
    private static final int DEFAULT_INIT_TIMEOUT = 1000;
    private static final int DEFAULT_RETRIES = 4;
    private static final String INIT_TIMEOUT = "com.sun.jndi.dns.timeout.initial";
    private static final String RETRIES = "com.sun.jndi.dns.timeout.retries";
    private CT lookupCT;
    private static final String LOOKUP_ATTR = "com.sun.jndi.dns.lookup.attr";
    private static final String RECURSION = "com.sun.jndi.dns.recursion";
    private static final int ANY = 255;
    private static final ZoneNode zoneTree;
    private static final boolean debug = false;
    
    public DnsContext(String str, String[] array, Hashtable<?, ?> hashtable) throws NamingException {
        this.domain = new DnsName(str.endsWith(".") ? str : (str + "."));
        this.servers = ((array == null) ? null : ((String[])array.clone()));
        this.environment = (Hashtable)hashtable.clone();
        this.envShared = false;
        this.parentIsDns = false;
        this.resolver = null;
        this.initFromEnvironment();
    }
    
    DnsContext(DnsContext dnsContext, DnsName domain) {
        this(dnsContext);
        this.domain = domain;
        this.parentIsDns = true;
    }
    
    private DnsContext(DnsContext dnsContext) {
        this.environment = dnsContext.environment;
        boolean b = true;
        dnsContext.envShared = b;
        this.envShared = b;
        this.parentIsDns = dnsContext.parentIsDns;
        this.domain = dnsContext.domain;
        this.servers = dnsContext.servers;
        this.resolver = dnsContext.resolver;
        this.authoritative = dnsContext.authoritative;
        this.recursion = dnsContext.recursion;
        this.timeout = dnsContext.timeout;
        this.retries = dnsContext.retries;
        this.lookupCT = dnsContext.lookupCT;
    }
    
    @Override
    public void close() {
        if (this.resolver != null) {
            this.resolver.close();
            this.resolver = null;
        }
    }
    
    @Override
    protected Hashtable<?, ?> p_getEnvironment() {
        return this.environment;
    }
    
    @Override
    public Hashtable<?, ?> getEnvironment() throws NamingException {
        return (Hashtable)this.environment.clone();
    }
    
    @Override
    public Object addToEnvironment(String key, Object o) throws NamingException {
        if (key.equals("com.sun.jndi.dns.lookup.attr")) {
            this.lookupCT = this.getLookupCT((String)o);
        }
        else if (key.equals("java.naming.authoritative")) {
            this.authoritative = "true".equalsIgnoreCase((String)o);
        }
        else if (key.equals("com.sun.jndi.dns.recursion")) {
            this.recursion = "true".equalsIgnoreCase((String)o);
        }
        else if (key.equals("com.sun.jndi.dns.timeout.initial")) {
            int int2 = Integer.parseInt((String)o);
            if (this.timeout != int2) {
                this.timeout = int2;
                this.resolver = null;
            }
        }
        else if (key.equals("com.sun.jndi.dns.timeout.retries")) {
            int int1 = Integer.parseInt((String)o);
            if (this.retries != int1) {
                this.retries = int1;
                this.resolver = null;
            }
        }
        if (!this.envShared) {
            return this.environment.put(key, o);
        }
        if (this.environment.get(key) != o) {
            this.environment = (Hashtable)this.environment.clone();
            this.envShared = false;
            return this.environment.put(key, o);
        }
        return o;
    }
    
    @Override
    public Object removeFromEnvironment(String key) throws NamingException {
        if (key.equals("com.sun.jndi.dns.lookup.attr")) {
            this.lookupCT = this.getLookupCT(null);
        }
        else if (key.equals("java.naming.authoritative")) {
            this.authoritative = false;
        }
        else if (key.equals("com.sun.jndi.dns.recursion")) {
            this.recursion = true;
        }
        else if (key.equals("com.sun.jndi.dns.timeout.initial")) {
            if (this.timeout != 1000) {
                this.timeout = 1000;
                this.resolver = null;
            }
        }
        else if (key.equals("com.sun.jndi.dns.timeout.retries") && this.retries != 4) {
            this.retries = 4;
            this.resolver = null;
        }
        if (!this.envShared) {
            return this.environment.remove(key);
        }
        if (this.environment.get(key) != null) {
            this.environment = (Hashtable)this.environment.clone();
            this.envShared = false;
            return this.environment.remove(key);
        }
        return null;
    }
    
    void setProviderUrl(String value) {
        this.environment.put("java.naming.provider.url", value);
    }
    
    private void initFromEnvironment() throws InvalidAttributeIdentifierException {
        this.lookupCT = this.getLookupCT((String)this.environment.get("com.sun.jndi.dns.lookup.attr"));
        this.authoritative = "true".equalsIgnoreCase((String)this.environment.get("java.naming.authoritative"));
        String anotherString = (String)this.environment.get("com.sun.jndi.dns.recursion");
        this.recursion = (anotherString == null || "true".equalsIgnoreCase(anotherString));
        String s = (String)this.environment.get("com.sun.jndi.dns.timeout.initial");
        this.timeout = ((s == null) ? 1000 : Integer.parseInt(s));
        String s2 = (String)this.environment.get("com.sun.jndi.dns.timeout.retries");
        this.retries = ((s2 == null) ? 4 : Integer.parseInt(s2));
    }
    
    private CT getLookupCT(String s) throws InvalidAttributeIdentifierException {
        return (s == null) ? new CT(1, 16) : fromAttrId(s);
    }
    
    public Object c_lookup(Name remain, Continuation continuation) throws NamingException {
        continuation.setSuccess();
        if (remain.isEmpty()) {
            DnsContext dnsContext = new DnsContext(this);
            dnsContext.resolver = new com.sun.jndi.dns.Resolver(this.servers, this.timeout, this.retries);
            return dnsContext;
        }
        try {
            DnsName fullyQualify = this.fullyQualify(remain);
            return DirectoryManager.getObjectInstance(new DnsContext(this, fullyQualify), remain, this, this.environment, rrsToAttrs(this.getResolver().query(fullyQualify, this.lookupCT.rrclass, this.lookupCT.rrtype, this.recursion, this.authoritative), (CT[])null));
        }
        catch (NamingException e) {
            continuation.setError(this, remain);
            throw continuation.fillInException(e);
        }
        catch (Exception rootCause) {
            continuation.setError(this, remain);
            NamingException e2 = new NamingException("Problem generating object using object factory");
            e2.setRootCause(rootCause);
            throw continuation.fillInException(e2);
        }
    }
    
    public Object c_lookupLink(Name name, Continuation continuation) throws NamingException {
        return this.c_lookup(name, continuation);
    }
    
    public NamingEnumeration<NameClassPair> c_list(Name remain, Continuation continuation) throws NamingException {
        continuation.setSuccess();
        try {
            DnsName fullyQualify = this.fullyQualify(remain);
            return new NameClassPairEnumeration(new DnsContext(this, fullyQualify), this.getNameNode(fullyQualify).getChildren());
        }
        catch (NamingException e) {
            continuation.setError(this, remain);
            throw continuation.fillInException(e);
        }
    }
    
    public NamingEnumeration<Binding> c_listBindings(Name remain, Continuation continuation) throws NamingException {
        continuation.setSuccess();
        try {
            DnsName fullyQualify = this.fullyQualify(remain);
            return new BindingEnumeration(new DnsContext(this, fullyQualify), this.getNameNode(fullyQualify).getChildren());
        }
        catch (NamingException e) {
            continuation.setError(this, remain);
            throw continuation.fillInException(e);
        }
    }
    
    public void c_bind(Name remain, Object o, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public void c_rebind(Name remain, Object o, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public void c_unbind(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public void c_rename(Name remain, Name name, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public Context c_createSubcontext(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public void c_destroySubcontext(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public NameParser c_getNameParser(Name name, Continuation continuation) throws NamingException {
        continuation.setSuccess();
        return DnsContext.nameParser;
    }
    
    public void c_bind(Name remain, Object o, Attributes attributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public void c_rebind(Name remain, Object o, Attributes attributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public DirContext c_createSubcontext(Name remain, Attributes attributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public Attributes c_getAttributes(Name remain, String[] array, Continuation continuation) throws NamingException {
        continuation.setSuccess();
        try {
            DnsName fullyQualify = this.fullyQualify(remain);
            CT[] attrIdsToClassesAndTypes = attrIdsToClassesAndTypes(array);
            CT classAndTypeToQuery = getClassAndTypeToQuery(attrIdsToClassesAndTypes);
            return rrsToAttrs(this.getResolver().query(fullyQualify, classAndTypeToQuery.rrclass, classAndTypeToQuery.rrtype, this.recursion, this.authoritative), attrIdsToClassesAndTypes);
        }
        catch (NamingException e) {
            continuation.setError(this, remain);
            throw continuation.fillInException(e);
        }
    }
    
    public void c_modifyAttributes(Name remain, int n, Attributes attributes, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public void c_modifyAttributes(Name remain, ModificationItem[] array, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public NamingEnumeration<SearchResult> c_search(Name name, Attributes attributes, String[] array, Continuation continuation) throws NamingException {
        throw new OperationNotSupportedException();
    }
    
    public NamingEnumeration<SearchResult> c_search(Name name, String s, SearchControls searchControls, Continuation continuation) throws NamingException {
        throw new OperationNotSupportedException();
    }
    
    public NamingEnumeration<SearchResult> c_search(Name name, String s, Object[] array, SearchControls searchControls, Continuation continuation) throws NamingException {
        throw new OperationNotSupportedException();
    }
    
    public DirContext c_getSchema(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    public DirContext c_getSchemaClassDefinition(Name remain, Continuation continuation) throws NamingException {
        continuation.setError(this, remain);
        throw continuation.fillInException(new OperationNotSupportedException());
    }
    
    @Override
    public String getNameInNamespace() {
        return this.domain.toString();
    }
    
    @Override
    public Name composeName(Name addAll, Name addAll2) throws NamingException {
        if (!(addAll2 instanceof DnsName) && !(addAll2 instanceof CompositeName)) {
            addAll2 = new DnsName().addAll(addAll2);
        }
        if (!(addAll instanceof DnsName) && !(addAll instanceof CompositeName)) {
            addAll = new DnsName().addAll(addAll);
        }
        if (addAll2 instanceof DnsName && addAll instanceof DnsName) {
            DnsName dnsName = (DnsName)(DnsName)addAll2.clone();
            dnsName.addAll(addAll);
            return new CompositeName().add(dnsName.toString());
        }
        Name prefix = (addAll2 instanceof CompositeName) ? addAll2 : new CompositeName().add(addAll2.toString());
        Name name = (addAll instanceof CompositeName) ? addAll : new CompositeName().add(addAll.toString());
        int n = prefix.size() - 1;
        if (name.isEmpty() || name.get(0).equals("") || prefix.isEmpty() || prefix.get(n).equals("")) {
            return super.composeName(name, prefix);
        }
        Name name2 = (addAll2 == prefix) ? ((CompositeName)prefix.clone()) : prefix;
        name2.addAll(name);
        if (this.parentIsDns) {
            Object o = (addAll2 instanceof DnsName) ? ((Object)addAll2.clone()) : new DnsName(prefix.get(n));
            o.addAll((addAll instanceof DnsName) ? addAll : new DnsName(name.get(0)));
            name2.remove(n + 1);
            name2.remove(n);
            name2.add(n, o.toString());
            return name2;
        }
        return name2;
    }
    
    private synchronized com.sun.jndi.dns.Resolver getResolver() throws NamingException {
        if (this.resolver == null) {
            this.resolver = new com.sun.jndi.dns.Resolver(this.servers, this.timeout, this.retries);
        }
        return this.resolver;
    }
    
    DnsName fullyQualify(Name name) throws NamingException {
        if (name.isEmpty()) {
            return this.domain;
        }
        DnsName obj = (name instanceof CompositeName) ? new DnsName(name.get(0)) : ((DnsName)new DnsName().addAll(name));
        if (obj.hasRootLabel()) {
            if (this.domain.size() == 1) {
                return obj;
            }
            throw new InvalidNameException("DNS name " + obj + " not relative to " + this.domain);
        }
        else {
            return (DnsName)obj.addAll(0, this.domain);
        }
    }
    
    private static Attributes rrsToAttrs(ResourceRecords resourceRecords, CT[] array) {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        for (int i = 0; i < resourceRecords.answer.size(); ++i) {
            ResourceRecord resourceRecord = (ResourceRecord)resourceRecords.answer.elementAt(i);
            int type = resourceRecord.getType();
            int rrclass = resourceRecord.getRrclass();
            if (classAndTypeMatch(rrclass, type, array)) {
                String attrId = toAttrId(rrclass, type);
                Attribute value = basicAttributes.get(attrId);
                if (value == null) {
                    value = new BasicAttribute(attrId);
                    basicAttributes.put(value);
                }
                value.add(resourceRecord.getRdata());
            }
        }
        return basicAttributes;
    }
    
    private static boolean classAndTypeMatch(int n, int n2, CT[] array) {
        if (array == null) {
            return true;
        }
        int i = 0;
        while (i < array.length) {
            CT ct = array[i];
            int n3 = ct.rrclass == 255 || ct.rrclass == n;
            int n4 = ct.rrtype == 255 || ct.rrtype == n2;
            if (n3 && n4) {
                return true;
            }
            ++i;
        }
        return false;
    }
    
    private static String toAttrId(int n, int n2) {
        String str = ResourceRecord.getTypeName(n2);
        if (n != 1) {
            str = ResourceRecord.getRrclassName(n) + " " + str;
            return str;
        }
        return str;
    }
    
    private static CT fromAttrId(String s) throws InvalidAttributeIdentifierException {
        if (s.equals("")) {
            throw new InvalidAttributeIdentifierException("Attribute ID cannot be empty");
        }
        int index = s.indexOf(32);
        int rrclass;
        if (index < 0) {
            rrclass = 1;
        }
        else {
            String substring = s.substring(0, index);
            rrclass = ResourceRecord.getRrclass(substring);
            if (rrclass < 0) {
                throw new InvalidAttributeIdentifierException("Unknown resource record class '" + substring + 39);
            }
        }
        String substring2 = s.substring(index + 1);
        int type = ResourceRecord.getType(substring2);
        if (type < 0) {
            throw new InvalidAttributeIdentifierException("Unknown resource record type '" + substring2 + 39);
        }
        return new CT(rrclass, type);
    }
    
    private static CT[] attrIdsToClassesAndTypes(String[] array) throws InvalidAttributeIdentifierException {
        if (array == null) {
            return null;
        }
        CT[] array2 = new CT[array.length];
        for (int i = 0; i < array.length; ++i) {
            array2[i] = fromAttrId(array[i]);
        }
        return array2;
    }
    
    private static CT getClassAndTypeToQuery(CT[] array) {
        int rrclass;
        int rrtype;
        if (array == null) {
            rrclass = 255;
            rrtype = 255;
        }
        else if (array.length == 0) {
            rrclass = 1;
            rrtype = 255;
        }
        else {
            rrclass = array[0].rrclass;
            rrtype = array[0].rrtype;
            for (int i = 1; i < array.length; ++i) {
                if (rrclass != array[i].rrclass) {
                    rrclass = 255;
                }
                if (rrtype != array[i].rrtype) {
                    rrtype = 255;
                }
            }
        }
        return new CT(rrclass, rrtype);
    }
    
    private NameNode getNameNode(DnsName obj) throws NamingException {
        dprint("getNameNode(" + obj + ")");
        ZoneNode deepestPopulated;
        synchronized (DnsContext.zoneTree) {
            deepestPopulated = DnsContext.zoneTree.getDeepestPopulated(obj);
        }
        dprint("Deepest related zone in zone tree: " + ((deepestPopulated != null) ? deepestPopulated.getLabel() : "[none]"));
        if (deepestPopulated != null) {
            NameNode contents;
            synchronized (deepestPopulated) {
                contents = deepestPopulated.getContents();
            }
            if (contents != null) {
                NameNode value = contents.get(obj, deepestPopulated.depth() + 1);
                if (value != null && !value.isZoneCut()) {
                    dprint("Found node " + obj + " in zone tree");
                    int zoneCurrent = this.isZoneCurrent(deepestPopulated, (DnsName)obj.getPrefix(deepestPopulated.depth() + 1));
                    boolean n = false;
                    synchronized (deepestPopulated) {
                        if (contents != deepestPopulated.getContents()) {
                            n = true;
                        }
                        else if (zoneCurrent == 0) {
                            deepestPopulated.depopulate();
                        }
                        else {
                            return value;
                        }
                    }
                    dprint("Zone not current; discarding node");
                    if (!(n == 0)) {
                        return this.getNameNode(obj);
                    }
                }
            }
        }
        dprint("Adding node " + obj + " to zone tree");
        DnsName zoneName = this.getResolver().findZoneName(obj, 1, this.recursion);
        dprint("Node's zone is " + zoneName);
        synchronized (DnsContext.zoneTree) {
            deepestPopulated = (ZoneNode)DnsContext.zoneTree.add(zoneName, 1);
        }
        NameNode contents;
        synchronized (deepestPopulated) {
            contents = (deepestPopulated.isPopulated() ? deepestPopulated.getContents() : this.populateZone(deepestPopulated, zoneName));
        }
        NameNode value2 = contents.get(obj, zoneName.size());
        if (value2 == null) {
            throw new ConfigurationException("DNS error: node not found in its own zone");
        }
        dprint("Found node in newly-populated zone");
        return value2;
    }
    
    private NameNode populateZone(ZoneNode zoneNode, DnsName obj) throws NamingException {
        dprint("Populating zone " + obj);
        ResourceRecords queryZone = this.getResolver().queryZone(obj, 1, this.recursion);
        dprint("zone xfer complete: " + queryZone.answer.size() + " records");
        return zoneNode.populate(obj, queryZone);
    }
    
    private boolean isZoneCurrent(ZoneNode zoneNode, DnsName dnsName) throws NamingException {
        if (!zoneNode.isPopulated()) {
            return false;
        }
        ResourceRecord soa = this.getResolver().findSoa(dnsName, 1, this.recursion);
        synchronized (zoneNode) {
            if (soa == null) {
                zoneNode.depopulate();
            }
            return zoneNode.isPopulated() && zoneNode.compareSerialNumberTo(soa) >= 0;
        }
    }
    
    private static final void dprint(String s) {
    }
    
    static {
        nameParser = new DnsNameParser();
        zoneTree = new ZoneNode(null);
    }
}
